"""Module for testing spiders."""


import json
import tempfile

from pathlib import Path

from finscraper.spiders import ISArticle, ILArticle


def test_ISArticle():
    # Test scraping
    spider = ISArticle().scrape(10)
    df = spider.get()
    assert len(df) >= 10
    assert len(df.columns) == 8

    # Test continuing scraping
    df2 = spider.scrape(10).get()
    assert len(df2) > len(df)

    # Save and load spider
    jobdir = spider.save()
    spider = ISArticle.load(jobdir)

    df3 = spider.scrape(10).get()
    assert len(df3) > len(df2)


def test_ILArticle():
    # Test scraping
    spider = ILArticle().scrape(10)
    df = spider.get()
    assert len(df) >= 10
    assert len(df.columns) == 8

    # Test continuing scraping
    df2 = spider.scrape(10).get()
    assert len(df2) > len(df)

    # Save and load spider
    jobdir = spider.save()
    spider = ILArticle.load(jobdir)

    df3 = spider.scrape(10).get()
    assert len(df3) > len(df2)


def test_spider_save_load_with_jobdir():
    jobdir = '../jobdir'
    category = 'jaakiekko'

    spider = ISArticle(category=category, jobdir=jobdir)
    
    save_jobdir = spider.save()
    loaded_spider = ISArticle.load(save_jobdir)

    assert (jobdir == str(spider.jobdir)
            == save_jobdir == str(loaded_spider.jobdir))
    assert category == spider.category == loaded_spider.category


def test_spider_save_load_without_jobdir():
    category = 'jaakiekko'

    spider = ISArticle(category=category)
    
    save_jobdir = spider.save()
    loaded_spider = ISArticle.load(save_jobdir)

    assert spider.jobdir is not None
    assert (str(spider.jobdir) == save_jobdir == str(loaded_spider.jobdir))
    assert category == spider.category == loaded_spider.category


def test_spider_clear():
    # Directory not removed when deleted
    spider = ISArticle()
    jobdir = spider.jobdir
    del spider
    assert Path(jobdir).exists()

    # Directory contents removed when cleared
    spider = ISArticle().scrape(1)
    jobdir = spider.save()
    spider_save_path = spider.spider_save_path
    items_save_path = spider.items_save_path
    assert Path(spider_save_path).exists()
    assert Path(items_save_path).exists()
    spider.clear()
    assert not Path(spider_save_path).exists()
    assert not Path(items_save_path).exists()



"""Module for spiders."""


from textwrap import indent

from finscraper.wrappers import _SpiderWrapper
from finscraper.scrapy_spiders.ilarticle import _ILArticleSpider, \
    _ILArticleItem
from finscraper.scrapy_spiders.isarticle import _ISArticleSpider, \
    _ISArticleItem


__jobdir_doc__ = '''
jobdir (None or str, optional): Working directory of the spider.
    Defaults to None, which creates a temp directory to be used.
    Note that this directory will only be deleted through the
    `clear` method!
'''


def _get_docstring(spider_cls, item_cls):
    return ( spider_cls.__init__.__doc__.strip()
            + indent(__jobdir_doc__, ' ' * 12)
            + indent(item_cls.__doc__, ' ' * 4))


class ISArticle(_SpiderWrapper):
    __doc__ = _get_docstring(_ISArticleSpider, _ISArticleItem)
    def __init__(self, category=None, follow_link_extractor=None,
                 item_link_extractor=None, jobdir=None):
        super(ISArticle, self).__init__(
            spider_cls=_ISArticleSpider,
            spider_params=dict(
                category=category,
                follow_link_extractor=follow_link_extractor,
                item_link_extractor=item_link_extractor
            ),
            jobdir=jobdir)


class ILArticle(_SpiderWrapper):
    __doc__ = _get_docstring(_ILArticleSpider, _ILArticleItem)
    def __init__(self, category=None, follow_link_extractor=None,
                 item_link_extractor=None, jobdir=None):
        super(ILArticle, self).__init__(
            spider_cls=_ILArticleSpider,
            spider_params=dict(
                category=category,
                follow_link_extractor=follow_link_extractor,
                item_link_extractor=item_link_extractor
            ),
            jobdir=jobdir)


"""Module for wrapping Scrapy spiders."""


import json
import pickle
import shutil
import tempfile
import uuid

from multiprocessing import Process, Queue
from pathlib import Path

import pandas as pd

from twisted.internet import reactor

from scrapy import Request
from scrapy.crawler import CrawlerProcess, CrawlerRunner
from scrapy.exceptions import CloseSpider
from scrapy.settings import Settings
from scrapy.spiders import Spider
from scrapy.utils.log import configure_logging


def _run_func_as_process(func, *args, **kwargs):
    q = Queue()
    p = Process(target=func, args=(q, *args),
                kwargs=kwargs)
    p.start()
    result = q.get()
    p.join()

    if result is not None:
        raise result


def _run_spider_func(q, spider_cls, spider_params, settings):
    try:
        configure_logging()
        runner = CrawlerRunner(settings)
        deferred = runner.crawl(spider_cls, **spider_params)
        deferred.addBoth(lambda _: reactor.stop())
        reactor.run()
        q.put(None)
    except Exception as e:
        Queue.put(e)


class _SpiderWrapper:

    def __init__(self, spider_cls, spider_params, jobdir=None):
        self.spider_cls = spider_cls
        self.spider_params = spider_params

        if jobdir is None:
            self._jobdir = Path(tempfile.gettempdir()) / str(uuid.uuid4())
        else:
            self._jobdir = Path(jobdir)
        self._jobdir.mkdir(parents=True, exist_ok=True)
        
        self._items_save_path = self._jobdir / 'items.jl'
        self._spider_save_path = self._jobdir / 'spider.pkl'
        
        # Note: Parameters cannot be changed outside by setting them
        for param in self.spider_params:
            setattr(self, param, self.spider_params[param])

    @property
    def jobdir(self):
        return str(self._jobdir)

    @property
    def items_save_path(self):
        return str(self._items_save_path)

    @property
    def spider_save_path(self):
        return str(self._spider_save_path)

    def _run_spider(self, itemcount=10, timeout=0, pagecount=0, errorcount=0,
                    settings=None):
        settings_ = Settings()
        settings_.setmodule('finscraper.settings', priority='project')
        settings_['JOBDIR'] = self.jobdir
        settings_['FEEDS'] = {self.items_save_path: {'format': 'jsonlines'}}
        settings_['CLOSESPIDER_ITEMCOUNT'] = itemcount
        settings_['CLOSESPIDER_TIMEOUT'] = timeout
        settings_['CLOSESPIDER_PAGECOUNT'] = pagecount
        settings_['CLOSESPIDER_ERRORCOUNT'] = errorcount
        if settings is not None:
            settings_.update(settings)
        try:
            _run_func_as_process(
                func=_run_spider_func,
                spider_cls=self.spider_cls,
                spider_params=self.spider_params,
                settings=settings_
            )
        except KeyboardInterrupt:
            pass
    
    def scrape(self, n=10, settings=None):
        """Scrape given number of items.
        
        Args:
            n (int, optional): Number of items to scrape. Zero corresponds to
                no limit. Defaults to 10.
            settings (dict or None, optional): Scrapy spider settings to use.
                Defaults to None, which correspond to default settings.
                See list of available settings at:
                https://docs.scrapy.org/en/latest/topics/settings.html.
        
        Returns:
            self
        """
        self._run_spider(itemcount=n, settings=settings)
        return self

    def get(self, fmt='df'):
        """Return scraped data as DataFrame or list.
        
        Args:
            fmt (str, optional): Format to return parsed items as. Should be
                in ['df', 'list']. Defaults to 'df'.

        Returns:
            DataFrame, when fmt = 'df'.
            List of scraped items, when fmt = 'list'.
        """
        if fmt not in ['df', 'list']:
            ValueError(f'Format {fmt} should be in ["df", "list"]')
        jsonlines = []
        if self._items_save_path.exists():
            with open(self.items_save_path, 'r') as f:
                for line in f:
                    jsonlines.append(json.loads(line))
        if fmt == 'list':
            return jsonlines
        elif fmt == 'df':
            return pd.DataFrame(jsonlines)
    
    def save(self):
        """Save spider in `jobdir` for later use.
        
        Returns:
            Path to job directory as a string.
        """
        save_tuple = (self.spider_cls, self.spider_params, self.jobdir)
        with open(self.spider_save_path, 'wb') as f:
            pickle.dump(save_tuple, f)
        return self.jobdir

    @classmethod
    def load(cls, jobdir):
        """Load spider from `jobdir`."""
        expected_path = Path(jobdir) / 'spider.pkl'
        with open(expected_path, 'rb') as f:
            (spider_cls, spider_params, jobdir) = pickle.load(f)
        return cls(jobdir=jobdir, **spider_params)

    def clear(self):
        """Clear contents of `jobdir`."""
        if self._jobdir.exists():
            shutil.rmtree(self._jobdir)
        self._jobdir.mkdir(parents=True, exist_ok=True)



"""Module for ILArticle spider."""


import time

from scrapy import Item, Field, Selector
from scrapy.crawler import Spider
from scrapy.linkextractors import LinkExtractor
from scrapy.loader import ItemLoader
from scrapy.loader.processors import TakeFirst, Identity, MapCompose

from finscraper.scrapy_spiders.mixins import FollowAndParseItemMixin
from finscraper.utils import strip_join


class _ILArticleSpider(FollowAndParseItemMixin, Spider):
    name = 'ilarticle'
    custom_settings = {}

    def __init__(
            self,
            category=None,
            follow_link_extractor=None,
            item_link_extractor=None,
            *args,
            **kwargs):
        """Fetch Iltalehti news articles.
        
        Args:
            category (str, list or None, optional): Category to fetch articles
                from, meaning pages under https://iltalehti.fi/<category>/*'.
                Defaults to None, which fetches articles everywhere.
            follow_link_extractor (scrapy.linkextractors.LinkExtractor or
                None, optional): Link extractor to use for finding new article
                pages. Defaults to None, which uses the default follow link
                extractor.
            item_link_extractor (scrapy.linkextractors.LinkExtractor or
                None, optional): Link extractor for fetching article pages to
                scrape. Defaults to None, which uses the default item link
                extractor.
        """
        super(_ILArticleSpider, self).__init__(*args, **kwargs)
        self.category = category
        self.follow_link_extractor = follow_link_extractor
        self.item_link_extractor = item_link_extractor

        article_suffix = r'.*/a/[0-9A-z\-]+'
        if category is None:
            self.start_urls = ['https://www.iltalehti.fi']
            self.allow_follow = ()
            self.allow_items = (article_suffix)
        else:
            if type(category) == str:
                category = [category]
            self.start_urls = []
            self.allow_follow = []
            self.allow_items = []
            for cat in category:
                self.start_urls.append(f'https://www.iltalehti.fi/{cat}')
                self.allow_follow.append(rf'.*/{cat}/.*')
                self.allow_items.append(rf'.*/{cat}/{article_suffix}')

        self.allow_domains = ('iltalehti.fi')
        self.deny_domains = ()
        self.deny = ()

        if self.follow_link_extractor is None:
            self.follow_link_extractor = LinkExtractor(
                allow_domains=self.allow_domains,
                allow=self.allow_follow,
                deny=self.deny,
                deny_domains=self.deny_domains,
                canonicalize=True
            )
        if self.item_link_extractor is None:
            self.item_link_extractor = LinkExtractor(
                allow_domains=self.allow_domains,
                allow=self.allow_items,
                deny=self.deny,
                deny_domains=self.deny_domains,
                canonicalize=True
            )

    @staticmethod
    def _get_image_metadata(text):
        sel = Selector(text=text)
        return {
            'src': sel.xpath('//img//@src').get(),
            'alt': sel.xpath('//img//@alt').get(),
            'caption': sel.xpath(
                '//div[contains(@class, "media-caption")]//text()').getall()
        }
    
    def _parse_item(self, resp):
        l = ItemLoader(item=_ILArticleItem(), response=resp)
        l.add_value('url', resp.url)
        l.add_value('time', int(time.time()))
        l.add_xpath('title',
            '//article//h1[contains(@class, "article-headline")]//text()')
        l.add_xpath('ingress',
            '//article//div[contains(@class, "article-description")]//text()')
        l.add_xpath('content',
            '//article//div[contains(@class, "article-body")]//text()')
        l.add_xpath('published',
            '//time//text()')
        l.add_xpath('author',
            '//article//div[contains(@class, "author-name")]//text()')
        l.add_xpath('images',
            '//article//div[contains(@class, "article-image")]')
        return l.load_item()


class _ILArticleItem(Item):
    """
    Returned fields:
        url (str): URL of the scraped web page.
        time (int): UNIX timestamp of the scraping.
        title (str): Title of the article.
        ingress (str): Ingress of the article.
        content (str): Contents of the article.
        published (str): Publish time of the article.
        author (str): Author of the article.
        images (list of dict): Images of the article.
    """
    url = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    time = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    title = Field(
        input_processor=strip_join,
        output_processor=TakeFirst()
    )
    ingress = Field(
        input_processor=strip_join,
        output_processor=TakeFirst()
    )
    content = Field(
        input_processor=strip_join,
        output_processor=TakeFirst()
    )
    published = Field(
        input_processor=strip_join,
        output_processor=TakeFirst()
    )
    author = Field(
        input_processor=strip_join,
        output_processor=TakeFirst()
    )
    images = Field(
        input_processor=MapCompose(_ILArticleSpider._get_image_metadata),
        output_processor=Identity()
    )



"""Module for Scrapy spider mixins."""


from scrapy import Request


class FollowAndParseItemMixin:
    """Parse items and follow links based on defined link extractors.
    
    When using this mixin, the following needs to be defined when inheriting:
        1) `item_link_extractor` attribute: Extract links to parse items from.
        2) `follow_link_extractor` attribute: Extract links to follow and find
            item pages from.
        3) `parse_item` function: Parse item from response.
    """
    item_link_extractor = None
    follow_link_extractor = None

    def _parse_item(self, resp):
        raise NotImplementedError('Function `_parse_item` not implemented!')

    def parse(self, resp, to_parse=False):
        """Parse items and follow links based on defined link extractors."""
        if to_parse:
            yield self._parse_item(resp)

        # Parse items and further on extract links from those pages
        item_links = self.item_link_extractor.extract_links(resp)
        for link in item_links:
            yield Request(link.url, callback=self.parse,
                          cb_kwargs={'to_parse': True})

        # Extract all links from this page
        follow_links = self.follow_link_extractor.extract_links(resp)
        for link in follow_links:
            yield Request(link.url, callback=self.parse)



import setuptools

with open('README.md', 'r') as f:
    long_description = f.read()

setuptools.setup(
    name='finscraper',
    version='0.0.1c',
    license='MIT',
    description='Web scraping API for Finnish websites',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Jesse Myrberg',
    author_email='jesse.myrberg@gmail.com',
    url='https://github.com/jmyrberg/finscraper',
    keywords=['web', 'scraping', 'finnish'],
    install_requires=[
        'pandas>=1.0.3',
        'scrapy>=2.1.0'
    ],
    packages=setuptools.find_packages(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3'
    ]
)